# -*- coding: utf-8 -*-
"""
:copyright: (c) 2015 - 2017 Stichting Koppeltaal
:license: AGPL, see `LICENSE.md` for more details.
"""

import functools
import hamcrest
import json
import pkg_resources
import six

from hamcrest.core.base_matcher import BaseMatcher
from koppeltaal import interfaces, transport
from six.moves.urllib.parse import urlparse, urlunparse, urlencode


unicode = six.text_type


class Response(transport.Response):

    def __init__(self, request_method=None, json=None, location=None):
        self.request_method = request_method
        self.json = json
        self.location = location


class MockTransport(object):

    def __init__(self, module_name):
        self._module_name = module_name
        self.clear()

    def _load(self, fixture):
        if fixture is None:
            return None
        filename = pkg_resources.resource_filename(self._module_name, fixture)
        with open(filename) as fp:
            return json.load(fp)

    def _expect(self, method, fixture, url):
        location = fixture.get('redirect_to')
        if 'respond_error' in fixture:
            raise interfaces.ResponseError(
                Response(
                    request_method=method,
                    json=self._load(fixture['respond_error']),
                    location=location))
        return Response(
            request_method=method,
            json=self._load(fixture.get('respond_with')),
            location=location)

    def clear(self):
        self.expected = {}
        self.called = {}

    def expect(self, method, url, **fixture):
        """Register an URL that the transport expects to be called for.

        Fixture(s) is the contents that the transport returns when the
        corresponding expected URL is called. When registering fixtures for
        the same URL multiple times, they are returned in that order.

        Note how the data that is to be sent to the expected URL is stored in
        the `called` attribute when indeed the expected URL is called.
        """
        expect = functools.partial(self._expect, method, fixture)
        self.expected.setdefault(url, []).append(expect)

    def relative_url(self, url, params=None):
        parts = list(map(unicode, urlparse(url)[2:]))
        url = urlunparse([unicode(''), unicode('')] + parts)
        if params:
            url += '?' + urlencode(params)
        return url

    def absolute_url(self, url):
        return url

    def query(self, url, params=None):
        url = self.relative_url(url, params)
        if not len(self.expected.get(url, [])):
            raise AssertionError('Unexpected url call', url)
        expect = self.expected[url].pop(0)
        response = expect(url)
        if response.request_method != 'GET':
            raise AssertionError(
                'Incorrect request method {} '
                'should be GET'.format(response.request_method))
        return response

    def query_redirect(self, url, params=None):
        url = self.relative_url(url, params)
        if not len(self.expected.get(url, [])):
            raise AssertionError('Unexpected url call', url)
        expect = self.expected[url].pop(0)
        response = expect(url)
        if response.request_method != 'GET':
            raise AssertionError(
                'Incorrect request method {} '
                'should be GET'.format(response.request_method))
        return response

    def create(self, url, data):
        url = self.relative_url(url)
        self.called[url] = data
        if not len(self.expected.get(url, [])):
            raise AssertionError('Unexpected url call', url)
        expect = self.expected[url].pop(0)
        response = expect(url)
        if response.request_method != 'POST':
            raise AssertionError(
                'Incorrect request method {} '
                'should be POST'.format(response.request_method))
        return response

    def update(self, url, data):
        url = self.relative_url(url)
        self.called[url] = data
        if not len(self.expected.get(url, [])):
            raise AssertionError('Unexpected url call', url)
        expect = self.expected[url].pop(0)
        response = expect(url)
        if response.request_method != 'PUT':
            raise AssertionError(
                'Incorrect request method {} '
                'should be PUT'.format(response.request_method))
        return response

    def close(self):
        pass


class HasFHIRResource(BaseMatcher):

    def __init__(self, resourcetype, containing=None):
        self.resourcetype = resourcetype
        self.containing = containing
        if containing is not None:
            self.matcher = hamcrest.has_entry(
                'entry',
                hamcrest.has_item(
                    hamcrest.has_entry(
                        'content',
                        hamcrest.all_of(
                            hamcrest.has_entry(
                                'resourceType', self.resourcetype),
                            self.containing))))
        else:
            self.matcher = hamcrest.has_entry(
                'entry',
                hamcrest.has_item(
                    hamcrest.has_entry(
                        'content',
                        hamcrest.has_entry(
                            'resourceType', self.resourcetype))))

    def _matches(self, json):
        return self.matcher.matches(json)

    def describe_to(self, description):
        description.append_text(
            'a FHIR resource entry of type {} '.format(self.resourcetype))
        if self.containing is not None:
            description.append_text('containing ')
            self.containing.describe_to(description)


has_resource = HasFHIRResource


class HasFHIRExtension(BaseMatcher):

    def __init__(self, url, containing=None):
        self.url = url
        self.containing = containing
        if containing is not None:
            self.matcher = hamcrest.has_entry(
                'extension',
                hamcrest.has_item(
                    hamcrest.all_of(
                        hamcrest.has_entry(
                            'url', hamcrest.ends_with(self.url)),
                        self.containing)))
        else:
            self.matcher = hamcrest.has_entry(
                'extension',
                hamcrest.has_item(
                    hamcrest.has_entry(
                        'url', hamcrest.ends_with(self.url))))

    def _matches(self, json):
        return self.matcher.matches(json)

    def describe_to(self, description):
        description.append_text(
            'a FHIR extension ending with {} '.format(self.url))
        if self.containing is not None:
            description.append_text('containing ')
            self.containing.describe_to(description)


has_extension = HasFHIRExtension


class HasProcessingStatusException(BaseMatcher):

    def __init__(self, error, containing=None):
        self.error = error
        self.containing = containing
        if containing is not None:
            self.matcher = has_extension(
                '#ProcessingStatus',
                hamcrest.all_of(
                    has_extension(
                        '#ProcessingStatusStatus',
                        hamcrest.has_entry('valueCode', 'Failed')),
                    has_extension(
                        '#ProcessingStatusException',
                        hamcrest.has_entry(
                            'valueString', self.error)),
                        self.containing))
        else:
            self.matcher = has_extension(
                '#ProcessingStatus',
                hamcrest.all_of(
                    has_extension(
                        '#ProcessingStatusStatus',
                        hamcrest.has_entry('valueCode', 'Failed')),
                    has_extension(
                        '#ProcessingStatusException',
                        hamcrest.has_entry(
                            'valueString', self.error))))

    def _matches(self, json):
        return self.matcher.matches(json)

    def describe_to(self, description):
        description.append_text(
            'a valueString {} '.format(self.error))
        if self.containing is not None:
            description.append_text('containing ')
            self.containing.describe_to(description)


has_exception = HasProcessingStatusException



# -*- coding: utf-8 -*-
"""
:copyright: (c) 2015 - 2017 Stichting Koppeltaal
:license: AGPL, see `LICENSE.md` for more details.
"""

import json

from koppeltaal.fhir import packer
from koppeltaal import (
    codes,
    fhir,
    interfaces,
    utils)


MARKER = object()


class Entry(object):
    _model = MARKER
    _content = MARKER
    _definition = MARKER
    _resource_type = MARKER
    _standard_type = MARKER
    _fhir_link = MARKER
    _history_less_fhir_link = MARKER

    def __init__(self, packer, content=None, model=None):
        self._packer = packer

        if content is not None:
            self._content = content.copy()
            resource_type = self._content.get('resourceType', 'Other')
            if resource_type == 'Other':
                self._standard_type = False
                assert 'code' in self._content
                for code in self._content.get('code').get('coding', []):
                    resource_type = codes.OTHER_RESOURCE_USAGE.unpack_coding(
                        code)
            assert resource_type != 'Other'
            self._resource_type = resource_type
            self._definition = fhir.REGISTRY.definition_for_type(
                self.resource_type)
        if model is not None:
            self._model = model

    @property
    def definition(self):
        if self._definition is not MARKER:
            return self._definition
        assert self._model is not MARKER, 'Need a model.'
        self._definition = fhir.REGISTRY.definition_for_model(self._model)
        return self._definition

    @property
    def resource_type(self):
        if self._resource_type is not MARKER:
            return self._resource_type
        self._resource_type, self._standard_type = \
            fhir.REGISTRY.type_for_definition(self.definition)
        return self._resource_type

    @property
    def standard_type(self):
        if self._standard_type is not MARKER:
            return self._standard_type
        self._resource_type, self._standard_type = \
            fhir.REGISTRY.type_for_definition(self.definition)
        return self._standard_type

    @property
    def fhir_link(self):
        if self._fhir_link is not MARKER:
            return self._fhir_link
        assert self._model is not MARKER, 'Need a model.'

        if self._model is None:
            return None

        if self._model.fhir_link is not None:
            self._fhir_link = self._model.fhir_link
            return self._fhir_link

        if interfaces.IIdentifiedFHIRResource.providedBy(self._model):
            self._fhir_link = self._packer.fhir_link(
                self._model, self.resource_type)
            self._model.fhir_link = self._fhir_link
            return self._fhir_link

        return None

    @property
    def history_less_fhir_link(self):
        if self._history_less_fhir_link is not MARKER:
            return self._history_less_fhir_link

        link = self.fhir_link
        if link is not None:
            self._history_less_fhir_link = utils.strip_history_from_link(link)
            return self._history_less_fhir_link

        return None

    def unpack(self):
        if self._model is not MARKER:
            return self._model

        self._model = None
        if self.definition is not None:
            self._model = self._packer.unpack(
                self._content, self.definition, allow_broken=True)
            if self._model is not None:
                self._model.fhir_link = self.fhir_link
        return self._model

    def pack(self):
        if self._content is MARKER:
            if self.definition is None:
                raise interfaces.InvalidResource(None, self._model)
            self._content = self._packer.pack(self._model, self.definition)
            if self.standard_type:
                self._content['resourceType'] = self.resource_type
            else:
                # This is not a standard fhir resource type.
                self._content['resourceType'] = 'Other'
                self._content['code'] = {
                    'coding': [
                        codes.OTHER_RESOURCE_USAGE.pack_coding(
                            self.resource_type)]}
        return self._content

    def __eq__(self, other):
        if isinstance(other, dict) and 'reference' in other:
            return other['reference'] == self.fhir_link
        if interfaces.IFHIRResource.providedBy(other):
            if self._model is not MARKER:
                return self._model is other
            assert self.fhir_link is not None, 'Should not happen'
            return other.fhir_link == self.fhir_link
        return NotImplemented

    def __format__(self, _):
        return '<{} fhir_link="{}" type="{}">{}</{}>'.format(
            self.__class__.__name__,
            self.fhir_link,
            self.resource_type,
            json.dumps(self._content, indent=2, sort_keys=True),
            self.__class__.__name__)


class Resource(object):
    _create_entry = Entry

    def __init__(self, domain=None, integration=None):
        self.items = []
        self.domain = domain
        self.integration = integration
        self.packer = packer.Packer(self, integration.fhir_link)

    def add_payload(self, response):
        self.items.append(self._create_entry(self.packer, content=response))

    def add_model(self, model):
        assert interfaces.IFHIRResource.providedBy(model), \
            'Can only add resources'
        entry = self.find(model)
        if entry is None:
            entry = self._create_entry(self.packer, model=model)
            self.items.append(entry)
        return entry

    def find(self, entry):
        for item in self.items:
            if entry == item:
                # Entry provides a smart "comparison".
                return item
        return None

    def pack(self):
        for item in self.items:
            yield item.pack()

    def get_payload(self):
        for payload in self.pack():
            return payload

    def unpack(self):
        for item in self.items:
            yield item.unpack()

    def unpack_model(self, definition):
        expected_model = None
        for model in self.unpack():
            if definition.providedBy(model):
                if expected_model is not None:
                    # We allow only one expected_model header in the bundle.
                    raise interfaces.InvalidBundle(self)
                expected_model = model
        return expected_model



# -*- coding: utf-8 -*-
"""
:copyright: (c) 2015 - 2017 Stichting Koppeltaal
:license: AGPL, see `LICENSE.md` for more details.
"""

import hamcrest
import zope.interface.verify
import koppeltaal.definitions
import koppeltaal.interfaces
import koppeltaal.testing


def test_recieve_careteam(connector, transport):
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?'
        '_query=MessageHeader.GetNextNewAndClaim',
        respond_with='fixtures/bundle_one_message.json')
    transport.expect(
        'GET',
        '/FHIR/Koppeltaal/MessageHeader/_search?'
        '_query=MessageHeader.GetNextNewAndClaim',
        respond_with='fixtures/bundle_zero_messages.json')
    transport.expect(
        'PUT',
        '/FHIR/Koppeltaal/MessageHeader/45909'
        '/_history/2016-07-15T11:50:24:494.7839',
        respond_with='fixtures/resource_put_message.json')

    updates = list(connector.updates())
    assert len(updates) == 1
    for update in updates:
        with update:
            assert update.message.event == 'CreateOrUpdateCarePlan'
            careplan = update.data
            patient = update.patient
            assert zope.interface.verify.verifyObject(
                koppeltaal.definitions.CarePlan, update.data)

    response = transport.called.get(
        '/FHIR/Koppeltaal/MessageHeader/45909'
        '/_history/2016-07-15T11:50:24:494.7839')
    assert response is not None
    hamcrest.assert_that(
        response,
        koppeltaal.testing.has_extension(
            '#ProcessingStatus',
            koppeltaal.testing.has_extension(
                '#ProcessingStatusStatus',
                hamcrest.has_entry('valueCode', 'Success'))))

    assert len(careplan.participants) == 3
    for participant in careplan.participants:
        assert zope.interface.verify.verifyObject(
            koppeltaal.definitions.Participant, participant)
        assert len(participant.careteam) == 2
        assert participant.careteam[0].name == 'The Therapists Team'
        assert participant.careteam[1].name == 'The Supervisors Team'
        assert participant.careteam[0].subject is patient



# -*- coding: utf-8 -*-
"""
:copyright: (c) 2015 - 2017 Stichting Koppeltaal
:license: AGPL, see `LICENSE.md` for more details.
"""

import koppeltaal.interfaces
import koppeltaal.definitions


class Code(dict):

    def __init__(self, system, items):
        super(Code, self).__init__(items)
        if not system.startswith('http:'):
            system = koppeltaal.interfaces.NAMESPACE + system
        self.system = system

    def pack_code(self, value):
        if value not in self:
            raise koppeltaal.interfaces.InvalidCode(self, value)
        return value

    def pack_coding(self, value):
        if value not in self:
            raise koppeltaal.interfaces.InvalidCode(self, value)
        display = self[value]
        coding = {"code": value, "system": self.system}
        if display is not None:
            coding["display"] = display
        return coding

    def unpack_code(self, code):
        if code not in self:
            raise koppeltaal.interfaces.InvalidCode(self, code)
        return code

    def unpack_coding(self, coding):
        value = coding.get("code")
        system = coding.get("system")
        if system == koppeltaal.definitions.NULL_SYSTEM \
                and value == koppeltaal.definitions.NULL_VALUE:
            return None
        if system != self.system:
            raise koppeltaal.interfaces.InvalidSystem(self, system)
        if value not in self:
            raise koppeltaal.interfaces.InvalidCode(self, value)
        return value


ACTIVITY_KIND = Code(
    'ActivityKind',
    {'ELearning': 'E-Learning',
     'Game': 'Game',
     'Meeting': 'Meeting',
     'MultipleActivityTemplate': 'Multiple activity template',
     'Questionnaire': 'Questionnaire',
     })

ACTIVITY_LAUNCH_TYPE = Code(
    'ActivityDefinitionLaunchType',
    {'Mobile': 'Mobile',
     'None': 'None',
     'Web': 'Web',
     })

ACTIVITY_PERFORMER = Code(
    'ActivityPerformer',
    {'Practitioner': 'Practitioner',
     'RelatedPerson': 'RelatedPerson',
     'Patient': 'Patient',
     })

ADDRESS_USE = Code(
    'http://hl7.org/fhir/address-use',
    {'home': 'home',
     'old': 'old',
     'temp': 'temp',
     'work': 'work',
     })

CAREPLAN_ACTIVITY_CATEGORY = Code(
    'http://hl7.org/fhir/care-plan-activity-category',
    {'diet': 'diet',
     'drug': 'drug',
     'encounter': 'encounter',
     'observation': 'observation',
     'other': 'other',
     'procedure': 'procedure',
     'supply': 'supply',
     })

CAREPLAN_PARTICIPANT_ROLE = Code(
    'CarePlanParticipantRole',
    {'Analyst': 'Analyst',
     'Caregiver': 'Caregiver',
     'Secretary': 'Secretary',
     'Supervisor': 'Supervisor',
     'Thirdparty': 'Thirdparty',
     'Requester': 'Requester',
     })

CAREPLAN_ACTIVITY_STATUS = Code(
    'CarePlanActivityStatus',
    {'Available': 'Available',
     'Cancelled': 'Cancelled',
     'Completed': 'Completed',
     'Expired': 'Expired',
     'InProgress': 'InProgress',
     'SkippedByUser': 'SkippedByUser',
     'Waiting': 'Waiting',
     })

CAREPLAN_GOAL_STATUS = Code(
    'http://hl7.org/fhir/care-plan-goal-status',
    {'achieved': 'achieved',
     'cancelled': 'cancelled',
     'in progress': 'in progress',
     'sustaining': 'sustaining',
     })

CAREPLAN_STATUS = Code(
    'http://hl7.org/fhir/care-plan-status',
    {'active': 'active',
     'completed': 'completed',
     'planned': 'planned',
     })

CONTACT_SYSTEM = Code(
    'http://hl7.org/fhir/contact-system',
    {'email': 'email',
     'fax': 'fax',
     'phone': 'phone',
     'url': 'url',
     })

CARE_TEAM_STATUS = Code(
    'CareTeamStatus',
    {'active': 'Active',
     'entered-in-error': 'Entered in error',
     'inactive': 'Inactive',
     'proposed': 'Proposed',
     'suspended': 'Suspended',
     })

CONTACT_ENTITY_TYPE = Code(
    'http://hl7.org/fhir/contactentity-type',
    {'ADMIN': 'ADMIN',
     'BILL': 'BILL',
     'HR': 'HR',
     'PATINF': 'PATINF',
     'PAYOR': 'PAYOR',
     'PRESS': 'PRESS',
     })

CONTACT_USE = Code(
    'http://hl7.org/fhir/contact-use',
    {'home': 'home',
     'mobile': 'mobile',
     'old': 'old',
     'temp': 'temp',
     'work': 'work',
     })

DEVICE_KIND = Code(
    'DeviceKind',
    {'Application': 'Application'})

GENDER = Code(
    'http://hl7.org/fhir/v3/AdministrativeGender',
    {'F': 'Female',
     'M': 'Male',
     'UN': 'Undifferentiated',
     })

IDENTIFIER_USE = Code(
    'http://hl7.org/fhir/identifier-use',
    {'official': 'official',
     'secondary': 'secondary',
     'temp': 'temp',
     'usual': 'usual',
     })

MESSAGE_HEADER_RESPONSE_CODE = Code(
    'http://hl7.org/fhir/response-code',
    {'fatal-error': 'fatal-error',
     'ok': 'ok',
     'transient-error': 'transient-error',
     })

MESSAGE_HEADER_EVENTS = Code(
    'MessageEvents',
    {'CreateOrUpdateActivityDefinition': 'CreateOrUpdateActivityDefinition',
     'CreateOrUpdateCarePlan': 'CreateOrUpdateCarePlan',
     'CreateOrUpdateCarePlanActivityResult':
        'CreateOrUpdateCarePlanActivityResult',
     'CreateOrUpdatePatient': 'CreateOrUpdatePatient',
     'CreateOrUpdatePractitioner': 'CreateOrUpdatePractitioner',
     'CreateOrUpdateRelatedPerson': 'CreateOrUpdateRelatedPerson',
     'CreateOrUpdateUserMessage': 'CreateOrUpdateUserMessage',
     'UpdateCarePlanActivityStatus': 'UpdateCarePlanActivityStatus',
     })

MESSAGE_KIND = Code(
    'UserMessageKind',
    {'Advice': 'Advice',
     'Alert': 'Alert',
     'Answer': 'Answer',
     'Message': 'Message',
     'Notification': 'Notification',
     'Question': 'Question',
     'Request': 'Request',
     })

NAME_USE = Code(
    'http://hl7.org/fhir/name-use',
    {'anonymous': 'anonymous',
     'maiden': 'maiden',
     'nickname': 'nickname',
     'official': 'official',
     'old': 'old',
     'temp': 'temp',
     'usual': 'usual',
     })

ORGANIZATION_TYPE = Code(
    'http://hl7.org/fhir/organization-type',
    {'dept': 'dept',
     'edu': 'edu',
     'fed': 'fed',
     'icu': 'icu',
     'ins': 'ins',
     'pharm': 'pharm',
     'reli': 'reli',
     'team': 'team',
     })

# BlackBoxState is not valid, but the javascript connector generate
# some of those.
OTHER_RESOURCE_USAGE = Code(
    'OtherResourceUsage',
    {'ActivityDefinition': 'ActivityDefinition',
     'BlackBoxState': 'BlackBoxState',
     'CarePlanActivityStatus': 'CarePlanActivityStatus',
     'CareTeam': 'CareTeam',
     'StorageItem': 'StorageItem',
     'UserMessage': 'UserMessage',
     })


PROCESSING_STATUS = Code(
    'ProcessingStatus',
    {'Claimed': 'Claimed',
     'Failed': 'Failed',
     'MaximumRetriesExceeded': 'MaximumRetriesExceeded',
     'New': 'New',
     'ReplacedByNewVersion': 'ReplacedByNewVersion',
     'Success': 'Success',
     })


ISSUE_TYPE = Code(
    'http://hl7.org/fhir/issue-type',
    {'invalid': None,
     'structure': None,
     'required': None,
     'value': None,
     'invariant': None,
     'security': None,
     'login': None,
     'unknown': None,
     'expired': None,
     'forbidden': None,
     'processing': None,
     'not-supported': None,
     'duplicate': None,
     'not-found': None,
     'too-long': None,
     'code-unknown': None,
     'extension': None,
     'too-costly': None,
     'business-rule': None,
     'conflict': None,
     'transient': None,
     'lock-error': None,
     'no-store': None,
     'exception': None,
     'timeout': None,
     'throttled': None,
     })


ISSUE_SEVERITY = Code(
    'http://hl7.org/fhir/issue-severity',
    {'fatal': None,
     'error': None,
     'warning': None,
     'information': None,
     })



# -*- coding: utf-8 -*-
"""
:copyright: (c) 2015 - 2017 Stichting Koppeltaal
:license: AGPL, see `LICENSE.md` for more details.
"""

import six
import zope.interface

from koppeltaal.fhir import bundle, resource
from koppeltaal import (
    interfaces,
    definitions,
    logger,
    models,
    transport,
    utils)
from six.moves.urllib.parse import urlencode


unicode = six.text_type


DEFAULT_COUNT = 100


@zope.interface.implementer(interfaces.IIntegration)
class Integration(object):

    def __init__(
            self,
            name='Generic python application',
            url='https://example.com/fhir/Koppeltaal'):
        self.name = name
        self.url = url

    def transaction_hook(self, commit_function, message):
        return commit_function(message)

    def model_id(self, model):
        # You should in your implementation extend this class and
        # re-implement this method so that model_id() of a given model
        # always return the exact same id. This is NOT done here in
        # this simplistic default implementation.
        return id(model)

    def fhir_link(self, model, resource_type):
        return '{}/{}/{}'.format(
            self.url, resource_type, self.model_id(model))


@zope.interface.implementer(interfaces.IUpdate)
class Update(object):

    def __init__(self, message, resources, ack_function):
        self.message = message
        self.resources = resources
        assert len(message.data) == 1, 'Multiple focal resources found'
        self.data = message.data[0]
        self.patient = message.patient
        self._ack_function = ack_function

    def __enter__(self):
        self.acked = False

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None and exc_val is None and exc_tb is None:
            if self.acked is False:
                self.success()
        else:
            # There was an exception. We put back the message to new
            # since we assume we can be solved after.
            self.ack('New')
        if self.acked is not None:
            self._ack_function(self.message)

    def ack(self, status, exception=None):
        self.acked = True
        if self.message.status is None:
            self.message.status = models.ProcessingStatus()
        self.message.status.status = status
        self.message.status.exception = exception
        self.message.status.last_changed = utils.now()

    def success(self):
        self.ack('Success')

    def fail(self, exception="FAILED"):
        self.ack('Failed', unicode(exception))

    def postpone(self):
        self.acked = None


@zope.interface.implementer(interfaces.IConnector)
class Connector(object):
    _create_transport = transport.Transport

    def __init__(self, credentials, integration):
        self._credentials = credentials
        self.transport = self._create_transport(
            self._credentials.url,
            self._credentials.username,
            self._credentials.password)
        self.domain = self._credentials.domain
        self.integration = integration

    def _fetch_bundle(self, url, params=None, batch_count=None):
        count = 0
        next_url = url
        next_params = params
        packaging = bundle.Bundle(self.domain, self.integration)
        while next_url:
            response = self.transport.query(next_url, next_params)
            packaging.add_payload(response.json)
            next_url = utils.json2links(response.json).get('next')
            next_params = None  # Parameters are already in the next link.
            count += 1
            if batch_count is not None and count >= batch_count:
                break
        return packaging

    def metadata(self):
        return self.transport.query(interfaces.METADATA_URL).json

    def activities(self, archived=False):
        params = {'code': 'ActivityDefinition'}
        if archived:
            params['includearchived'] = 'yes'
        return self._fetch_bundle(
            interfaces.ACTIVITY_DEFINITION_URL, params).unpack()

    def activity(self, identifier, archived=False):
        for activity in self.activities(archived=archived):
            if activity.identifier == identifier:
                return activity
        return None

    def send_activity(self, activity):
        packaging = resource.Resource(self.domain, self.integration)
        packaging.add_model(activity)
        payload = packaging.get_payload()
        if activity.fhir_link is not None:
            response = self.transport.update(activity.fhir_link, payload)
        else:
            response = self.transport.create(interfaces.OTHER_URL, payload)
        if response.location is None:
            raise interfaces.ResponseError(response)
        activity.fhir_link = response.location
        return activity

    def launch(self, careplan, user=None, activity_identifier=None):
        activity = None
        for candidate in careplan.activities:
            if (activity_identifier is None or
                    candidate.identifier == activity_identifier):
                activity = candidate
                break
        if activity is None:
            raise interfaces.KoppeltaalError('No activity found')
        if user is None:
            user = careplan.patient
        application_id = None
        if activity.definition is not None:
            activity_definition = self.activity(activity.definition)
            assert interfaces.IReferredFHIRResource.providedBy(
                activity_definition.application)
            application_id = activity_definition.application.display
        return self.launch_from_parameters(
            application_id,
            careplan.patient.fhir_link,
            user.fhir_link,
            activity.identifier)

    def launch_from_parameters(
            self,
            application_id,
            patient_link,
            user_link,
            activity_identifier):
        assert application_id is not None, 'Invalid activity'
        assert patient_link is not None, 'Invalid patient'
        assert user_link is not None, 'Invalid user'
        params = {
            'client_id': application_id,
            'patient': patient_link,
            'user': user_link,
            'resource': activity_identifier}
        return self.transport.query_redirect(
            interfaces.OAUTH_LAUNCH_URL, params).location

    def authorize_from_parameters(
            self,
            application_id,
            launch_id,
            redirect_uri):
        assert application_id is not None, 'Invalid activity'
        return '{}?{}'.format(
            self.transport.absolute_url(interfaces.OAUTH_AUTHORIZE_URL),
            urlencode(
                (('client_id', application_id),
                 ('redirect_uri', redirect_uri),
                 ('response_type', 'code'),
                 ('scope', 'patient/*.read launch:{}'.format(launch_id)))))

    def token_from_parameters(self, code, redirect_url):
        params = {
            'code': code,
            'grant_type': 'authorization_code',
            'redirect_uri': redirect_url}
        # Note how for this specific request a different set of credentials
        # ought to be used.
        # See https://www.koppeltaal.nl/
        #    wiki/Technical_Design_Document_Koppeltaal_1.1.1
        #    #6._Game_retrieves_an_Access_token
        username = self.integration.client_id
        password = self.integration.client_secret
        assert username is not None, 'client id missing'
        assert password is not None, 'client secret missing'
        return self.transport.query(
            interfaces.OAUTH_TOKEN_URL,
            params=params,
            username=username,
            password=password).json

    def updates(self, expected_events=None):

        def send_back(message):
            packaging = resource.Resource(self.domain, self.integration)
            packaging.add_model(message)
            self.transport.update(message.fhir_link, packaging.get_payload())

        def send_back_on_transaction(message):
            return self.integration.transaction_hook(send_back, message)

        p = {'_query': 'MessageHeader.GetNextNewAndClaim'}
        while True:
            try:
                bundle = self._fetch_bundle(interfaces.MESSAGE_HEADER_URL, p)
                message = bundle.unpack_model(definitions.MessageHeader)
            except interfaces.InvalidBundle as error:
                logger.error(
                    'Bundle error while reading message: {}'.format(error))
                continue
            except interfaces.TransportError as error:
                logger.error(
                    'Transport error while reading mailbox: {}'.format(error))
                break

            if message is None:
                # We are out of messages
                break

            update = Update(message, bundle.unpack, send_back_on_transaction)
            errors = bundle.errors()
            if errors:
                reason = u', '.join([str(error) for error in errors])
                logger.error(
                    "Error while reading message '{}': {}".format(
                        message.fhir_link, reason))
                with update:
                    update.fail(reason)
            elif (expected_events is not None
                  and message.event not in expected_events):
                logger.warning(
                    "Event '{}' not expected in message '{}'".format(
                        message.event, message.fhir_link))
                with update:
                    update.fail('Event not expected')
            elif (message.source is not None and
                  message.source.endpoint == self.integration.url):
                logger.info(
                    'Event "{}" originated from our endpoint '
                    '"{}"'.format(message.event, self.integration.url))
                with update:
                    # We are the sender ourselves. Ack those messages.
                    update.success()
            else:
                yield update

    def search(
            self, message_id=None, event=None, status=None, patient=None,
            batch_size=DEFAULT_COUNT, batch_count=None):
        params = {}
        if message_id:
            params['_id'] = message_id
        else:
            params['_summary'] = 'true'
            params['_count'] = batch_size
        if event:
            params['event'] = event
        if status:
            params['ProcessingStatus'] = status
        if patient:
            params['Patient'] = patient.fhir_link
        return self._fetch_bundle(
            interfaces.MESSAGE_HEADER_URL, params, batch_count).unpack()

    def send(self, event, data, patient=None):
        identifier = utils.messageid()
        source = models.MessageHeaderSource(
            name=unicode(self.integration.name),
            endpoint=unicode(self.integration.url),
            software=unicode(interfaces.SOFTWARE),
            version=unicode(interfaces.VERSION))
        request_message = models.MessageHeader(
            timestamp=utils.now(),
            event=event,
            identifier=identifier,
            data=[data],
            source=source,
            patient=patient)
        request_bundle = bundle.Bundle(self.domain, self.integration)
        request_bundle.add_model(request_message)
        request_payload = request_bundle.get_payload()

        try:
            response = self.transport.create(
                interfaces.MAILBOX_URL, request_payload)
        except interfaces.ResponseError as error:
            response_resource = resource.Resource(
                self.domain, self.integration)
            response_resource.add_payload(error.response.json)
            outcome = response_resource.unpack_model(
                definitions.OperationOutcome)
            raise interfaces.OperationOutcomeError(outcome)

        response_bundle = bundle.Bundle(self.domain, self.integration)
        response_bundle.add_payload(response.json)
        response_message = response_bundle.unpack_model(
            definitions.MessageHeader)
        if (response_message is None or
                response_message.event != event or
                response_message.response is None or
                response_message.response.identifier != identifier or
                response_message.response.code != "ok"):
            raise interfaces.MessageResponseError(response_message)
        return response_message.data

    def close(self):
        self.transport.close()


@zope.interface.implementer(interfaces.IConnector)
class DummyConnector(object):
    """A dummy connector follows the API of a connector but do not do
    anything.
    """
    _create_transport = None

    def __init__(self, credentials, integration):
        self.transport = None
        self.domain = credentials.domain
        self.integration = integration

    def metadata(self):
        return {
            'name': 'Koppeltaal',
            'version': 'v1.1',
            'fhirVersion': '0.0.82'}

    def activities(self, archived=False):
        return []

    def activity(self, identifier, archived=False):
        return None

    def send_activity(self, activity):
        raise interfaces.DummyError()

    def launch(self, careplan, user=None, activity_identifier=None):
        raise interfaces.DummyError()

    def launch_from_parameters(
            self,
            application_id,
            patient_link,
            user_link,
            activity_identifier):
        raise interfaces.DummyError()

    def authorize_from_parameters(
            self,
            application_id,
            launch_id,
            redirect_uri):
        raise interfaces.DummyError()

    def token_from_parameters(self, code, redirect_url):
        return {}

    def updates(self, expected_events=None):
        return []

    def search(
            self, message_id=None, event=None, status=None, patient=None):
        return None

    def send(self, event, data, patient=None):
        raise interfaces.DummyError()

    def close(self):
        pass



# -*- coding: utf-8 -*-
"""
:copyright: (c) 2015 - 2017 Stichting Koppeltaal
:license: AGPL, see `LICENSE.md` for more details.
"""

import datetime
import pytest
import zope.interface.verify
import koppeltaal.codes
import koppeltaal.definitions
import koppeltaal.fhir.packer
import koppeltaal.fhir.resource
import koppeltaal.interfaces
import koppeltaal.models
import koppeltaal.utils

import sys
if sys.version_info.major == 2:
    import mock

    def bytes(arg, encoding):
        return str(arg)

else:
    from unittest import mock


@pytest.fixture
def packer():
    return koppeltaal.fhir.packer.Packer(
        mock.MagicMock(
            **{'spec': koppeltaal.fhir.resource.Resource,
               'find.return_value': None}),
        mock.MagicMock(return_value='https://example.com/dummy/dummy'))


@pytest.fixture
def namespace():
    return koppeltaal.interfaces.NAMESPACE


def test_coding(packer, namespace):
    vertebrates = koppeltaal.codes.Code(
        'Vertebrate',
        {'amphibians': 'Amphibians',
         'birds': 'Birds',
         'mammals': 'Mammals',
         'human': None,
         'reptiles': 'Reptiles'})

    packed = vertebrates.pack_code('amphibians')
    assert packed == 'amphibians'

    packed = vertebrates.pack_code('human')
    assert packed == 'human'

    with pytest.raises(koppeltaal.interfaces.InvalidCode):
        vertebrates.pack_code('sponges')

    coding = vertebrates.pack_coding('amphibians')
    assert coding == {
        'code': 'amphibians',
        'display': 'Amphibians',
        'system': namespace + 'Vertebrate'}

    coding = vertebrates.pack_coding('human')
    assert coding == {
        'code': 'human',
        'system': namespace + 'Vertebrate'}

    with pytest.raises(koppeltaal.interfaces.InvalidCode):
        vertebrates.pack_coding('sponges')

    unpacked = vertebrates.unpack_code('amphibians')
    assert unpacked == 'amphibians'

    unpacked = vertebrates.unpack_code('human')
    assert unpacked == 'human'

    with pytest.raises(koppeltaal.interfaces.InvalidCode):
        vertebrates.unpack_code('sponges')

    unpacked_coding = vertebrates.unpack_coding({
        'code': 'amphibians',
        'display': 'Amphibians',
        'system': namespace + 'Vertebrate'})
    assert 'amphibians' == unpacked_coding

    unpacked_coding = vertebrates.unpack_coding({
        'code': 'human',
        'system': namespace + 'Vertebrate'})
    assert 'human' == unpacked_coding

    with pytest.raises(koppeltaal.interfaces.InvalidCode):
        vertebrates.unpack_coding({
            'code': 'sponges',
            'display': 'Sponges',
            'system': namespace + 'Vertebrate'})

    unknown_value = vertebrates.unpack_coding({
        'code': 'UNK',
        'display': 'Unknown',
        'system': koppeltaal.definitions.NULL_SYSTEM})
    assert unknown_value is None

    with pytest.raises(koppeltaal.interfaces.InvalidCode):
        vertebrates.unpack_coding({
            'code': 'UNKNOW',
            'display': 'Unknown',
            'system': koppeltaal.definitions.NULL_SYSTEM})

    with pytest.raises(koppeltaal.interfaces.InvalidSystem):
        vertebrates.unpack_coding({
            'code': 'reptiles',
            'display': 'Reptiles',
            'system': 'foobarbaz'})


def test_unpack_name(packer):
    name = packer.unpack(
        {'given': [u'Napoleon'],
         'family': [u'Bonaparte'],
         'use': u'official'},
        koppeltaal.definitions.Name)
    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.Name, name)
    assert name.given == [u'Napoleon']
    assert name.family == [u'Bonaparte']
    assert name.use == 'official'

    name = packer.unpack(
        {'given': [u'Thea'],
         'family': [u'van', u'de', u'Bovenburen'],
         'suffix': [u'van', u'den', u'Overkant'],
         'use': u'official'},
        koppeltaal.definitions.Name)
    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.Name, name)
    assert name.given == [u'Thea']
    assert name.family == [u'van', u'de', u'Bovenburen']
    assert name.suffix == [u'van', u'den', u'Overkant']
    assert name.use == 'official'

    name = packer.unpack(
        {'given': [u'Thea'],
         'family': [u'van', u'de', u'Bovenburen'],
         'suffix': [u'van', u'den', u'Overkant'],
         'text': u'Thea van de Bovenburen',
         'use': u'official'},
        koppeltaal.definitions.Name)
    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.Name, name)
    assert name.given == [u'Thea']
    assert name.family == [u'van', u'de', u'Bovenburen']
    assert name.suffix == [u'van', u'den', u'Overkant']
    assert name.text == 'Thea van de Bovenburen'
    assert name.use == 'official'

def test_pack_name(packer):
    name1 = packer.pack(
        koppeltaal.models.Name(
            given=[u'Napoleon'],
            family=[u'Bonaparte'],
            use='old'),
        koppeltaal.definitions.Name)

    assert name1 == {
        'family': ['Bonaparte'],
        'given': ['Napoleon'],
        'id': mock.ANY,
        'use': 'old'}

    name2 = packer.pack(
        koppeltaal.models.Name(
            given=[u'Nathan']),
        koppeltaal.definitions.Name)

    assert name2 == {
        'given': ['Nathan'],
        'id': mock.ANY,
        'use': 'official'}

    name3 = packer.pack(
        koppeltaal.models.Name(
            given=[u'Nathan'],
            family=[u'der', u'Fantasten'],
            suffix=[u'tot', u'Daarhelemaalië']),
        koppeltaal.definitions.Name)

    assert name3 == {
        'given': ['Nathan'],
        'family': [u'der', u'Fantasten'],
        'id': mock.ANY,
        'suffix': [u'tot', u'Daarhelemaalië'],
        'use': 'official'}

    # Verify an empty array value for (in this case) "given" does not leave an
    # empty array as that would trip up the Koppeltaal server's JSON parser.
    name4 = packer.pack(
        koppeltaal.models.Name(
            given=[],
            family=[u'der', u'Fantasten'],
            suffix=[u'tot', u'Daarhelemaalië']),
        koppeltaal.definitions.Name)

    assert name4 == {
        'family': [u'der', u'Fantasten'],
        'id': mock.ANY,
        'suffix': [u'tot', u'Daarhelemaalië'],
        'use': 'official'}

    # Verify a None value for (in this case) "given" does not leave an array
    # with null value(s) as that would trip up the Koppeltaal server's JSON
    # parser.
    with pytest.raises(koppeltaal.interfaces.InvalidValue):
        packer.pack(
            koppeltaal.models.Name(
                given=[None],
                family=[u'der', u'Fantasten'],
                suffix=[u'tot', u'Daarhelemaalië']),
            koppeltaal.definitions.Name)

    with pytest.raises(koppeltaal.interfaces.InvalidResource):
        packer.pack(
            'Napoleon',
            koppeltaal.definitions.Name)

    with pytest.raises(koppeltaal.interfaces.InvalidValue):
        packer.pack(
            koppeltaal.models.Name(
                given=[u'Napoleon'],
                family=[u'Bonaparte'],
                use='cool name'),
            koppeltaal.definitions.Name)

    name5 = packer.pack(
        koppeltaal.models.Name(
            given=[u'Nathan'],
            family=[u'der', u'Fantasten'],
            suffix=[u'tot', u'Daarhelemaalië'],
            text=u'Nathan de Fantasten tot Daarhelemaalië'),
        koppeltaal.definitions.Name)

    assert name5 == {
        'given': [u'Nathan'],
        'family': [u'der', u'Fantasten'],
        'id': mock.ANY,
        'suffix': [u'tot', u'Daarhelemaalië'],
        'text': u'Nathan de Fantasten tot Daarhelemaalië',
        'use': u'official'}

def test_unpack_patient(packer, namespace):
    patient1 = packer.unpack(
        {'active': True,
         'extension': [
             {'url': namespace + u'Patient#Age',
              'valueInteger': 42}],
         'gender': {'coding': [
             {'code': u'M',
              'display': u'M',
              'system': u'http://hl7.org/fhir/v3/AdministrativeGender'}]},
         'name': [
             {'given': [u'Paul'],
              'family': [u'Roger'],
              'text': u'Paul Roger',
              'use': u'official'}],
         'address': [{
            'city': u'Rotterdam',
            'country': u'The Netherlands',
            'period': {
                'start': u'2013-06-01T12:34:00',
                },
            'state': u'Zuid-Holland',
            'text': u'Rotterdam, ken je dat nie horen dan?',
            'use': u'work',
            'zip': u'3033CH'
          }]},
        koppeltaal.definitions.Patient)

    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.Patient, patient1)
    name = patient1.name[0]
    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.Name, name)
    assert name.given == [u'Paul']
    assert name.family == [u'Roger']
    assert name.text == u'Paul Roger'
    assert name.use == 'official'
    assert len(patient1.contacts) == 0
    assert len(patient1.identifiers) == 0
    assert patient1.age == 42
    assert patient1.active is True
    assert patient1.birth_date is None
    assert patient1.gender == 'M'

    address = patient1.address[0]
    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.Address, address)
    assert address.city == 'Rotterdam'
    assert address.country == 'The Netherlands'
    assert address.state == 'Zuid-Holland'
    assert address.text == 'Rotterdam, ken je dat nie horen dan?'
    assert address.use == 'work'
    assert address.zip == '3033CH'

    period = address.period
    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.Period, period)
    assert period.start == datetime.datetime(2013, 6, 1, 12, 34)
    assert period.end is None

    patient2 = packer.unpack(
        {'gender': {'coding': [
            {'code': u'UNK',
             'display': u'UNK',
             'system': u'http://hl7.org/fhir/v3/NullFlavor'}]},
         'name': [
             {'given': [u'Paul'],
              'family': [u'Roger'],
              'use': u'official'}]},
        koppeltaal.definitions.Patient)
    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.Patient, patient2)
    name2 = patient2.name[0]
    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.Name, name2)
    assert name2.given == [u'Paul']
    assert name2.family == [u'Roger']
    assert name2.use == 'official'
    assert len(patient2.contacts) == 0
    assert len(patient2.identifiers) == 0
    assert patient2.age is None
    assert patient2.active is None
    assert patient2.birth_date is None
    assert patient2.gender is None


def test_pack_patient(packer):
    patient1 = packer.pack(
        koppeltaal.models.Patient(
            active=True,
            age=42,
            name=[
                koppeltaal.models.Name(
                    given=[u'Paul'],
                    family=[u'Roger'],
                    use='official')],
            gender='M'),
        koppeltaal.definitions.Patient)

    assert patient1 == {
        'active': True,
        'extension': [
            {'url': 'http://ggz.koppeltaal.nl/fhir/Koppeltaal/Patient#Age',
             'valueInteger': 42}],
        'id': mock.ANY,
        'gender': {'coding': [
            {'code': 'M',
             'display': 'Male',
             'system': 'http://hl7.org/fhir/v3/AdministrativeGender'}]},
        'name': [
            {'id': mock.ANY,
             'given': ['Paul'],
             'family': ['Roger'],
             'use': 'official'}]}

    patient2 = packer.pack(
        koppeltaal.models.Patient(
            birth_date=datetime.datetime(1976, 6, 1, 12, 34),
            contacts=[
                koppeltaal.models.Contact(
                    system='email',
                    value=u'petra@example.com',
                    use='home')],
            identifiers=[
                koppeltaal.models.Identifier(
                    system=u'http://fhir.nl/fhir/NamingSystem/bsn',
                    value=u'640563569',
                    use='official')],
            name=[
                koppeltaal.models.Name(
                    given=[u'Petra'],
                    use='temp')],
            gender='F'),
        koppeltaal.definitions.Patient)

    assert patient2 == {
        'birthDate': '1976-06-01T12:34:00',
        'gender': {'coding': [
            {'code': 'F',
             'display': 'Female',
             'system': 'http://hl7.org/fhir/v3/AdministrativeGender'}]},
        'id': mock.ANY,
        'identifier': [
            {'id': mock.ANY,
             'system': 'http://fhir.nl/fhir/NamingSystem/bsn',
             'value': '640563569',
             'use': 'official'}],
        'name': [
            {'id': mock.ANY,
             'given': ['Petra'],
             'use': 'temp'}],
        'telecom': [
            {'id': mock.ANY,
             'system': 'email',
             'value': 'petra@example.com',
             'use': 'home'}]}


def test_unpack_practitioner(packer):
    practitioner1 = packer.unpack(
        {'telecom': [
            {'system': u'email',
             'value': u'paul@example.com',
             'use': u'work'}],
         'name':  {'given': [u'Paul'],
                   'family': [u'Cézanne'],
                   'text': u'Paul Cézanne',
                   'use': u'official'}},
        koppeltaal.definitions.Practitioner)

    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.Practitioner, practitioner1)
    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.Name, practitioner1.name)
    assert practitioner1.name.given == [u'Paul']
    assert practitioner1.name.family == [u'Cézanne']
    assert practitioner1.name.text == u'Paul Cézanne'
    assert practitioner1.name.use == 'official'
    assert len(practitioner1.contacts) == 1
    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.Contact, practitioner1.contacts[0])
    assert practitioner1.contacts[0].system == 'email'
    assert practitioner1.contacts[0].value == u'paul@example.com'
    assert practitioner1.contacts[0].use == 'work'
    assert len(practitioner1.identifiers) == 0


def test_pack_practitioner(packer):
    practitioner1 = packer.pack(
        koppeltaal.models.Practitioner(
            contacts=[
                koppeltaal.models.Contact(
                    system='email',
                    value=u'paul@example.com',
                    use='work')],
            identifiers=[
                koppeltaal.models.Identifier(
                    system=u'http://fhir.nl/fhir/NamingSystem/bsn',
                    value=u'154694496',
                    use='official')],
            name=koppeltaal.models.Name(
                given=[u'Paul'],
                family=[u'Cézanne'])),
        koppeltaal.definitions.Practitioner)

    assert practitioner1 == {
        'id': mock.ANY,
        'identifier': [
            {'id': mock.ANY,
             'system': 'http://fhir.nl/fhir/NamingSystem/bsn',
             'value': '154694496',
             'use': 'official'}],
        'name': {'id': mock.ANY,
                 'given': ['Paul'],
                 'family': [u'Cézanne'],
                 'use': 'official'},
        'telecom': [
            {'id': mock.ANY,
             'system': 'email',
             'value': 'paul@example.com',
             'use': 'work'}]}


def test_unpack_activity_definition(packer, namespace):
    definition1 = packer.unpack(
        {'extension': [
            {'url': (namespace +
                     u'ActivityDefinition#ActivityDefinitionIdentifier'),
             'valueString': u'myapp'},
            {'url': namespace + u'ActivityDefinition#ActivityName',
             'valueString': u'My APP'},
            {'url': namespace + u'ActivityDefinition#ActivityKind',
             'valueCoding': {
                 'code': u'Game',
                 'display': u'Game',
                 'system': namespace + u'ActivityKind'}},
            {'url': namespace + u'ActivityDefinition#Application',
             'valueResource': {
                 'reference': 'https://example.com/refmyapp',
                 'display': u'refmyapp'}}]},
        koppeltaal.definitions.ActivityDefinition)

    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.ActivityDefinition, definition1)
    assert definition1.identifier == 'myapp'
    assert definition1.name == 'My APP'
    assert definition1.kind == 'Game'
    assert zope.interface.verify.verifyObject(
        koppeltaal.interfaces.IReferredFHIRResource, definition1.application)
    assert definition1.application.fhir_link == 'https://example.com/refmyapp'
    assert definition1.application.display == 'refmyapp'
    assert definition1.launch_type == 'Web'
    assert definition1.is_active is True
    assert definition1.is_domain_specific is None
    assert definition1.is_archived is False
    assert len(definition1.subactivities) == 0


def test_unpack_message_header(packer, namespace):
    message1 = packer.unpack(
        {'data': [{'reference': 'https://example.com/data'}],
         'event': {'code': u'CreateOrUpdatePatient',
                   'display': u'CreateOrUpdatePatient',
                   'system': namespace + u'MessageEvents'},
         'identifier': u'42-42-42',
         'source': {'name': u'unpack message header',
                    'version': u'about 1.0',
                    'endpoint': u'https://example.com/here',
                    'software': u'pytest'},
         'timestamp': u'2015-10-09T12:16:00+00:00'},
        koppeltaal.definitions.MessageHeader)

    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.MessageHeader, message1)
    assert len(message1.data) == 1
    assert zope.interface.verify.verifyObject(
        koppeltaal.interfaces.IReferredFHIRResource, message1.data[0])
    assert message1.data[0].fhir_link == 'https://example.com/data'
    assert message1.event == 'CreateOrUpdatePatient'
    assert message1.identifier == '42-42-42'
    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.MessageHeaderSource, message1.source)
    assert message1.source.endpoint == 'https://example.com/here'
    assert message1.source.name == 'unpack message header'
    assert message1.source.software == 'pytest'
    assert message1.source.version == 'about 1.0'
    assert message1.timestamp == datetime.datetime(
        2015, 10, 9, 12, 16, tzinfo=koppeltaal.utils.utc)


def test_pack_message_header(packer, namespace):
    message1 = packer.pack(
        koppeltaal.models.MessageHeader(
            timestamp=datetime.datetime(2015, 10, 9, 12, 4),
            event='CreateOrUpdatePatient',
            identifier=u'42-42',
            source=koppeltaal.models.MessageHeaderSource(
                software=u"pytest",
                endpoint=u'https://example.com/here')),
        koppeltaal.definitions.MessageHeader)

    assert message1 == {
        'event': {'code': 'CreateOrUpdatePatient',
                  'display': 'CreateOrUpdatePatient',
                  'system': namespace + 'MessageEvents'},
        'id': mock.ANY,
        'identifier': u'42-42',
        'source': {'endpoint': u'https://example.com/here',
                   'id': mock.ANY,
                   'software': u'pytest'},
        'timestamp': '2015-10-09T12:04:00+00:00'}


def test_unpack_extension_invalid_integer(packer, namespace):
    with pytest.raises(koppeltaal.interfaces.InvalidValue) as error:
        packer.unpack(
            {'extension': [
                {'url': namespace + u'Patient#Age',
                 'valueInteger': u'forty'}],
             'name': [{'given': [u'Me']}]},
            koppeltaal.definitions.Patient)
    assert str(error.value) == \
        "InvalidValue: invalid value '{'url': " \
        "'http://ggz.koppeltaal.nl/fhir/Koppeltaal/Patient#Age', " \
        "'valueInteger': 'forty'}' for 'Patient.age'."

    with pytest.raises(koppeltaal.interfaces.InvalidValue) as error:
        packer.unpack(
            {'extension': [
                {'url': namespace + u'Patient#Age',
                 'value': 40}],
             'name': [{'given': [u'Me']}]},
            koppeltaal.definitions.Patient)
    assert str(error.value) == \
        "InvalidValue: invalid value '{'url': " \
        "'http://ggz.koppeltaal.nl/fhir/Koppeltaal/Patient#Age', " \
        "'value': 40}' for 'Patient.age'."


def test_unpack_extension_required_missing(packer, namespace):
    with pytest.raises(koppeltaal.interfaces.InvalidValue) as error:
        packer.unpack(
            {'extension': [
                {'url': (namespace +
                         u'ActivityDefinition#ActivityDefinitionIdentifier'),
                 'valueString': u'myapp'},
                {'url': namespace + u'ActivityDefinition#ActivityName',
                 'valueString': u'My APP'},
                {'url': namespace + u'ActivityDefinition#ActivityKind',
                 'valueCoding': {
                     'code': u'Game',
                     'display': u'Game',
                     'system': namespace + u'ActivityKind'}}]},
            koppeltaal.definitions.ActivityDefinition)
    assert str(error.value) == \
        "RequiredMissing: 'application' required but missing."


def test_unpack_native_invalid_boolean(packer):
    with pytest.raises(koppeltaal.interfaces.InvalidValue) as error:
        packer.unpack(
            {'active': u'yes!',
             'name': [{'given': [u'Me']}]},
            koppeltaal.definitions.Patient)
    assert str(error.value) == \
        "InvalidValue: invalid value 'yes!' for 'Patient.active'."


def test_unpack_native_invalid_code(packer):
    with pytest.raises(koppeltaal.interfaces.InvalidCode) as error:
        packer.unpack(
            {'use': u'cool name'},
            koppeltaal.definitions.Name)
    assert str(error.value) == \
        "InvalidCode: 'cool name' not in 'http://hl7.org/fhir/name-use'."

    with pytest.raises(koppeltaal.interfaces.InvalidValue) as error:
        packer.unpack(
            {'use': None},
            koppeltaal.definitions.Name)
    assert str(error.value) == \
        "InvalidValue: invalid value 'None' for 'Name.use'."


def test_unpack_native_invalid_coding(packer, namespace):
    with pytest.raises(koppeltaal.interfaces.InvalidCode) as error:
        packer.unpack(
            {'event': {'code': u'CreateWorld',
                       'display': u'CreateWorld',
                       'system': namespace + u'MessageEvents'},
             'identifier': u'42-42-42',
             'source': {'endpoint': u'https://example.com/here',
                        'software': u'pytest'},
             'timestamp': u'2015-10-09T12:16:00+00:00'},
            koppeltaal.definitions.MessageHeader)
    assert str(error.value) == \
        "InvalidCode: 'CreateWorld' not in " \
        "'http://ggz.koppeltaal.nl/fhir/Koppeltaal/MessageEvents'."

    with pytest.raises(koppeltaal.interfaces.InvalidCode) as error:
        packer.unpack(
            {'event': {'coding': u'CreateOrUpdatePatient',
                       'system': namespace + u'MessageEvents'},
             'identifier': u'42-42-42',
             'source': {'endpoint': u'https://example.com/here',
                        'software': u'pytest'},
             'timestamp': u'2015-10-09T12:16:00+00:00'},
            koppeltaal.definitions.MessageHeader)
    assert str(error.value) == \
        "InvalidCode: code is missing."

    with pytest.raises(koppeltaal.interfaces.InvalidSystem) as error:
        packer.unpack(
            {'event': {'code': u'CreateWorld',
                       'display': u'CreateWorld',
                       'system': u'http://example.com/event'},
             'identifier': u'42-42-42',
             'source': {'endpoint': u'https://example.com/here',
                        'software': u'pytest'},
             'timestamp': u'2015-10-09T12:16:00+00:00'},
            koppeltaal.definitions.MessageHeader)
    assert str(error.value) == \
        "InvalidSystem: system 'http://example.com/event' is not supported."

    with pytest.raises(koppeltaal.interfaces.InvalidSystem) as error:
        packer.unpack(
            {'event': {'code': u'CreateOrUpdatePatient'},
             'identifier': u'42-42-42',
             'source': {'endpoint': u'https://example.com/here',
                        'software': u'pytest'},
             'timestamp': u'2015-10-09T12:16:00+00:00'},
            koppeltaal.definitions.MessageHeader)
    assert str(error.value) == \
        "InvalidSystem: system is missing."

    with pytest.raises(koppeltaal.interfaces.InvalidValue) as error:
        packer.unpack(
            {'event': u'CreateOrUpdatePatient',
             'identifier': u'42-42-42',
             'source': {'endpoint': u'https://example.com/here',
                        'software': u'pytest'},
             'timestamp': u'2015-10-09T12:16:00+00:00'},
            koppeltaal.definitions.MessageHeader)
    assert str(error.value) == \
        "InvalidValue: invalid value 'CreateOrUpdatePatient' " \
        "for 'MessageHeader.event'."


def test_unpack_native_invalid_datetime(packer):
    with pytest.raises(koppeltaal.interfaces.InvalidValue) as error:
        packer.unpack(
            {'birthDate': u'yesterday',
             'name': [{'given': [u'Me']}]},
            koppeltaal.definitions.Patient)
    assert str(error.value) == \
        "InvalidValue: invalid value 'yesterday' for 'Patient.birth_date' " \
        "(FHIR name: 'birthDate')."

    with pytest.raises(koppeltaal.interfaces.InvalidValue) as error:
        packer.unpack(
            {'birthDate': -1,
             'name': [{'given': [u'Me']}]},
            koppeltaal.definitions.Patient)
    assert str(error.value) == \
        "InvalidValue: invalid value '-1' for 'Patient.birth_date' " \
        "(FHIR name: 'birthDate')."

    with pytest.raises(koppeltaal.interfaces.InvalidValue) as error:
        packer.unpack(
            {'birthDate': False,
             'name': [{'given': [u'Me']}]},
            koppeltaal.definitions.Patient)
    assert str(error.value) == \
        "InvalidValue: invalid value 'False' for 'Patient.birth_date' " \
        "(FHIR name: 'birthDate')."


def test_unpack_native_invalid_multiple(packer):
    with pytest.raises(koppeltaal.interfaces.InvalidValue) as error:
        packer.unpack(
            {'given': u'Napoleon'},
            koppeltaal.definitions.Name)
    assert str(error.value) == \
        "InvalidValue: invalid value 'Napoleon' for 'Name.given'."

    with pytest.raises(koppeltaal.interfaces.InvalidValue) as error:
        packer.unpack(
            {'family': 42},
            koppeltaal.definitions.Name)
    assert str(error.value) == \
        "InvalidValue: invalid value '42' for 'Name.family'."

    with pytest.raises(koppeltaal.interfaces.InvalidValue) as error:
        packer.unpack(
            {'family': [42]},
            koppeltaal.definitions.Name)
    assert str(error.value) == \
        "InvalidValue: invalid value '42' for 'Name.family'."


def test_unpack_native_invalid_object(packer):
    with pytest.raises(koppeltaal.interfaces.InvalidValue) as error:
        packer.unpack(
            {'telecom': u'by fax',
             'name': [{'given': [u'Me']}]},
            koppeltaal.definitions.Patient)
    assert str(error.value) == \
        "InvalidValue: invalid value 'by fax' for 'Patient.contacts' " \
        "(FHIR name: 'telecom')."

    with pytest.raises(koppeltaal.interfaces.InvalidValue) as error:
        packer.unpack(
            {'telecom': False,
             'name': [{'given': [u'Me']}]},
            koppeltaal.definitions.Patient)
    assert str(error.value) == \
        "InvalidValue: invalid value 'False' for 'Patient.contacts' " \
        "(FHIR name: 'telecom')."


def test_unpack_native_invalid_string(packer):
    with pytest.raises(koppeltaal.interfaces.InvalidValue) as error:
        packer.unpack(
            {'value': 42},
            koppeltaal.definitions.Identifier)
    assert str(error.value) == \
        "InvalidValue: invalid value '42' for 'Identifier.value'."

    with pytest.raises(koppeltaal.interfaces.InvalidValue) as error:
        packer.unpack(
            {'value': True},
            koppeltaal.definitions.Identifier)
    assert str(error.value) == \
        "InvalidValue: invalid value 'True' for 'Identifier.value'."


def test_unpack_allow_broken(packer):
    broken = packer.unpack(
        {'use': u'cool name'},
        koppeltaal.definitions.Name,
        allow_broken=True)
    assert not koppeltaal.definitions.Name.providedBy(broken)
    assert zope.interface.verify.verifyObject(
        koppeltaal.interfaces.IBrokenFHIRResource, broken)
    assert str(broken.error) == \
        "InvalidCode: 'cool name' not in 'http://hl7.org/fhir/name-use'."


def test_unpack_organization(packer, namespace):
    org1 = packer.unpack(
        {'active': True,
         'address': [{
             'city': u'Rotterdam',
             'country': u'The Netherlands',
             'period': {
                 'start': u'2013-06-01T12:34:00',
                 },
             'state': u'Zuid-Holland',
             'text': u'Rotterdam, ken je dat nie horen dan?',
             'use': u'work',
             'zip': u'3033CH'
             }, {
             'city': u'Den Haag',
             'country': u'The Netherlands',
             'period': {
                 'start': u'2010-06-01T12:34:00',
                 'end': u'2013-06-01T12:33:00',
                 },
             'state': u'Zuid-Holland',
             'text': u'Achter de duinuh',
             'use': u'work',
             'zip': u'2564TT'}]},
        koppeltaal.definitions.Organization)

    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.Organization, org1)

    addresses = org1.address
    assert 2 == len(addresses)

    address1 = addresses[0]
    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.Address, address1)
    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.Address, address1)
    assert address1.city == 'Rotterdam'
    assert address1.country == 'The Netherlands'
    assert address1.state == 'Zuid-Holland'
    assert address1.text == 'Rotterdam, ken je dat nie horen dan?'
    assert address1.use == 'work'
    assert address1.zip == '3033CH'

    period1 = address1.period
    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.Period, period1)
    assert period1.start == datetime.datetime(2013, 6, 1, 12, 34)
    assert period1.end is None

    address2 = addresses[1]
    period2 = address2.period
    assert zope.interface.verify.verifyObject(
        koppeltaal.definitions.Period, period2)
    assert period2.start == datetime.datetime(2010, 6, 1, 12, 34)
    assert period2.end == datetime.datetime(2013, 6, 1, 12, 33)


def test_pack_organization(packer):
    org1 = packer.pack(
        koppeltaal.models.Organization(
            active=True,
            address=[koppeltaal.models.Address(
                city=u'Rotterdam',
                country=u'The Netherlands',
                line=[u'Coolsingel 1'],
                period=koppeltaal.models.Period(
                    start=datetime.datetime(2010, 6, 1, 12, 34),
                    end=None),
                state=u'Zuid-Holland',
                text=u'Ken je dat nie horen dan?',
                use=u'work',
                zip=u'3030AB')],
            category=u'team',
            contacts=[],
            contact_persons=[],
            identifiers=[],
            name=u'Example',
            part_of=None),
        koppeltaal.definitions.Organization)

    assert org1 == {
        'active': True,
        'address': [{
            'city': 'Rotterdam',
            'country': 'The Netherlands',
            'id': 'ref002',
            'line': ['Coolsingel 1'],
            'period': {'id': 'ref001', 'start': '2010-06-01T12:34:00'},
            'state': 'Zuid-Holland',
            'text': 'Ken je dat nie horen dan?',
            'use': 'work',
            'zip': '3030AB'}],
        'id': 'ref003',
        'name': 'Example',
        'type': 'team'}



# -*- coding: utf-8 -*-
"""
:copyright: (c) 2015 - 2017 Stichting Koppeltaal
:license: AGPL, see `LICENSE.md` for more details.
"""

from koppeltaal.fhir.registry import Registry
from koppeltaal import definitions, models


REGISTRY = Registry({
    definitions.Activity: models.Activity,
    definitions.ActivityDefinition: models.ActivityDefinition,
    definitions.ActivityDetails: models.ActivityDetails,
    definitions.ActivityParticipant: models.Participant,
    definitions.Address: models.Address,
    definitions.CarePlan: models.CarePlan,
    definitions.CarePlanActivityStatus: models.ActivityStatus,
    definitions.CarePlanSubActivityStatus: models.SubActivity,
    definitions.CareTeam: models.CareTeam,
    definitions.Contact: models.Contact,
    definitions.Goal: models.Goal,
    definitions.Identifier: models.Identifier,
    definitions.Issue: models.Issue,
    definitions.MessageHeader: models.MessageHeader,
    definitions.MessageHeaderResponse: models.MessageHeaderResponse,
    definitions.MessageHeaderSource: models.MessageHeaderSource,
    definitions.Name: models.Name,
    definitions.OperationOutcome: models.OperationOutcome,
    definitions.Organization: models.Organization,
    definitions.OrganizationContactPerson: models.OrganizationContactPerson,
    definitions.Participant: models.Participant,
    definitions.Patient: models.Patient,
    definitions.Period: models.Period,
    definitions.Practitioner: models.Practitioner,
    definitions.ProcessingStatus: models.ProcessingStatus,
    definitions.SubActivity: models.SubActivity,
    definitions.SubActivityDefinition: models.SubActivityDefinition,
})



# -*- coding: utf-8 -*-
"""
:copyright: (c) 2015 - 2017 Stichting Koppeltaal
:license: AGPL, see `LICENSE.md` for more details.
"""

import zope.interface

from koppeltaal import definitions


def _inspect_definition(fields, definition):
    for name, field in definition.namesAndDescriptions():
        if not isinstance(field, definitions.Field):
            continue
        if field.extension:
            continue
        if field.multiple:
            fields.add(field.name)
        if field.field_type == 'object':
            _inspect_definition(fields, field.binding)


class Registry(dict):

    def repeatable_field_names(self, fhir_type):
        fields = {'extension', 'coding'}
        if fhir_type == 'Other':
            return fields

        for definition in self.keys():
            defined_type = definition.queryTaggedValue('resource type')
            if not defined_type:
                continue
            if defined_type[0] != fhir_type:
                continue
            _inspect_definition(fields, definition)

        return fields

    def definition_for_type(self, resource_type):
        definitions = []
        for definition in self.keys():
            defined_type = definition.queryTaggedValue('resource type')
            if defined_type and defined_type[0] == resource_type:
                definitions.append(definition)
        assert len(definitions) < 2, 'Too many definitions for resource type'
        if definitions:
            return definitions[0]
        return None

    def model_for_definition(self, definition):
        return self.get(definition)

    def definition_for_model(self, model):
        definitions = [
            d for d in zope.interface.providedBy(model).interfaces()
            if d in self]
        assert len(definitions) < 2, \
            'Too many definitions implemented by model'
        if definitions:
            return definitions[0]
        return None

    def type_for_definition(self, definition):
        assert definition in self, 'Unknown definition'
        return definition.queryTaggedValue('resource type')

    def type_for_model(self, model):
        definition = self.definition_for_model()
        if definition is None:
            return None
        return definition.queryTaggedValue('resource type')



# -*- coding: utf-8 -*-
"""
:copyright: (c) 2015 - 2017 Stichting Koppeltaal
:license: AGPL, see `LICENSE.md` for more details.
"""

import argparse
import json
import logging
import os
import os.path
import pdb
import sys
import dateutil

from koppeltaal import connector, codes, definitions, interfaces
from koppeltaal import models, logger, utils
from koppeltaal.fhir import xml, bundle


ACTIVITY_DEFINITION_OUTPUT = """Activity: {model.identifier}
- fhir link: {model.fhir_link}
- active: {model.is_active}
- application: {model.application.fhir_link}
- archived: {model.is_archived}
- description: {model.description}
- domain specific: {model.is_domain_specific}
- kind: {model.kind}
- name: {model.name}
- performer: {model.performer}
"""

ACTIVITY_STATUS_OUTPUT = """Activity Status: {model.identifier}
- fhir link: {model.fhir_link}
- status: {model.status}
"""

MESSAGE_OUTPUT = """Message: {model.identifier}
- fhir link: {model.fhir_link}
- event: {model.event}
- time stamp: {model.timestamp}
- software: {model.source.software} ({model.source.version})
- endpoint: {model.source.endpoint}
"""

CAREPLAN_OUTPUT = """CarePlan:
- fhir link: {model.fhir_link}
"""

PATIENT_OUTPUT = """Patient: {model.name}
- fhir link: {model.fhir_link}
"""

PRACTITIONER_OUTPUT = """Practitioner: {model.name}
- fhir link: {model.fhir_link}
"""

OUTPUT = {
    models.ActivityDefinition: ACTIVITY_DEFINITION_OUTPUT,
    models.ActivityStatus: ACTIVITY_STATUS_OUTPUT,
    models.CarePlan: CAREPLAN_OUTPUT,
    models.MessageHeader: MESSAGE_OUTPUT,
    models.Patient: PATIENT_OUTPUT,
    models.Practitioner: PRACTITIONER_OUTPUT,
}


def print_model(model):
    output = OUTPUT.get(model.__class__)
    if output:
        print(output.format(model=model))
    else:
        print(model)


def print_json(data):
    print(json.dumps(data, indent=2, sort_keys=True))


def get_credentials(args):
    # Domain is not required for all actions, so we're less strict about
    # requiring that.
    if args.username or args.password or args.domain:
        if not(args.username and args.password):
            sys.exit(
                'When supplying credentials through the commandline '
                'please always supply username and password.')
        else:
            return args.username, args.password, args.domain
    return utils.get_credentials_from_file(args.server)


class DummyResource(object):

    def __new__(cls, fhir_link):
        if fhir_link is None:
            return None
        return object.__new__(cls)

    def __init__(self, fhir_link):
        self.fhir_link = fhir_link


def directory(input):
    directory_name = os.path.abspath(os.path.expanduser(input))
    if not os.path.exists(directory_name):
        sys.exit(
            '"{}" does not exist'.format(input))
    if not os.path.isdir(directory_name):
        sys.exit(
            '"{}" is not a valid directory path'.format(input))
    return directory_name


def download(connection, directory, msgid=None, msg=None):
    if msg is not None:
        url = utils.strip_history_from_link(msg.fhir_link)
        msgid = url.rsplit('/', 1)[-1]
    response = connection.transport.query(
        interfaces.MESSAGE_HEADER_URL, {'_id': msgid}).json
    packaging = bundle.Bundle(connection.domain, connection.integration)
    packaging.add_payload(response)
    msg = packaging.unpack_model(definitions.MessageHeader)
    ts = msg.timestamp.isoformat()
    directory = os.path.join(directory, connection.domain)
    if not os.path.exists(directory):
        os.mkdir(directory)
    filename = os.path.join(directory, '{}-{}.json'.format(ts, msgid))
    with open(filename, 'w') as output:
        json.dump(
            response,
            output,
            indent=2,
            sort_keys=True)
    print('Wrote message "{}" to "{}"'.format(msgid, filename))


def _metadata(args, connection):
    print_json(connection.metadata())


def _validate(args, connection):
    payload = None
    if args.xml:
        payload = xml.xml2json(args.xml)
    if args.json:
        payload = json.load(args.json)
    if payload is None:
        print("Please provide an XML or JSON file.")
        return
    resource_bundle = bundle.Bundle('validation', connection.integration)
    resource_bundle.add_payload(payload)
    for model in resource_bundle.unpack():
        print_model(model)


def _activities(args, connection):
    for activity in connection.activities():
        print_model(activity)


def _messages(args, connection):
    for message in connection.search(
            event=args.event,
            status=args.status,
            patient=DummyResource(args.patient),
            batch_size=args.batch_size,
            batch_count=args.batch_count):
        if args.save_in_dir:
            download(connection, args.save_in_dir, msg=message)
        else:
            print_model(message)


def _message(args, connection):
    if args.save_in_dir:
        download(connection, args.save_in_dir, msgid=args.message_id)
    else:
        for model in connection.search(message_id=args.message_id):
            print_model(model)


def _updates(args, connection):
    until = None
    if args.until is not None:
        until = dateutil.parser.parse(args.until, tzinfos=utils.UTC)
    for index, update in enumerate(connection.updates()):
        with update:
            if until is None:
                if not args.all_updates and index:
                    update.postpone()
                    break
            else:
                if update.message.timestamp > until:
                    update.postpone()
                    break
            print_model(update.data)
            if args.failure:
                update.fail(args.failure)


def _launch(args, connection):
    print(
        connection.launch_from_parameters(
            args.application_id,
            args.patient_link,
            args.user_link,
            args.activity,
            ))


def console():
    parser = argparse.ArgumentParser(description='Koppeltaal connector')
    parser.add_argument('server', help='Koppeltaal server to connect to')
    parser.add_argument('--username')
    parser.add_argument('--password')
    parser.add_argument(
        '--domain',
        help='The domain on the server to send data to')
    parser.add_argument('--verbose', action='store_true')
    parser.add_argument(
        '--post-mortem',
        action='store_true',
        help='Debug any error with a Python debugger')

    subparsers = parser.add_subparsers(title='commands', dest='command')

    subparsers.add_parser('activities')
    subparsers.add_parser('metadata')

    validate = subparsers.add_parser('validate')
    validate.add_argument(
        '--xml', type=argparse.FileType('rb'),
        help="XML file to validate")
    validate.add_argument(
        '--json', type=argparse.FileType('rb'),
        help="JSON file to validate")

    messages = subparsers.add_parser('messages')
    messages.add_argument(
        '--patient',
        help='Patient FHIR link')
    messages.add_argument(
        '--status',
        choices=codes.PROCESSING_STATUS,
        help='Message header status')
    messages.add_argument(
        '--event', choices=codes.MESSAGE_HEADER_EVENTS,
        help='Event type')
    messages.add_argument(
        '--batch-size',
        type=int,
        help='Number of message per bundle.')
    messages.add_argument(
        '--batch-count',
        type=int,
        help='Number of bundles.')
    messages.add_argument(
        '--save-in-dir', type=directory,
        help='Save the source for each messsage listed in the query in '
             'the given directory')

    message = subparsers.add_parser('message')
    message.add_argument(
        'message_id',
        help='internal MessageHeader id or MessageHeader URL.')
    message.add_argument(
        '--save-in-dir', type=directory,
        help='Save the source for the messsage in the given directory')

    updates = subparsers.add_parser('updates')
    updates.add_argument(
        '--all', dest='all_updates', action='store_true',
        help='Process all next messages')
    updates.add_argument(
        '--until',
        help='Process all next messages until the given date (and time)')
    updates.add_argument(
        '--failure',
        help='Fail and set the exception on the messages.')

    launch = subparsers.add_parser('launch')
    launch.add_argument('user_link')
    launch.add_argument('application_id')
    launch.add_argument('patient_link')
    launch.add_argument('--activity', help='activity_identifier')

    args = parser.parse_args()

    root = logging.getLogger()
    root.addHandler(logging.StreamHandler(sys.stdout))

    if args.verbose:
        root.setLevel(logging.DEBUG)
        logger.set_log_level(logging.DEBUG)
    else:
        root.setLevel(logging.ERROR)
        logger.set_log_level(logging.ERROR)

    credentials = get_credentials(args)
    integration = connector.Integration(name='Python command line')
    connection = connector.Connector(credentials, integration)
    commands = {
        'activities': _activities,
        'launch': _launch,
        'message': _message,
        'messages': _messages,
        'metadata': _metadata,
        'updates': _updates,
        'validate': _validate}
    command = commands.get(args.command)
    if command is None:
        sys.exit('Unknown command {}'.format(args.command))
    try:
        command(args, connection)
    except Exception as error:
        if args.post_mortem:
            print(error)
            pdb.post_mortem()
        raise



# -*- coding: utf-8 -*-
"""
:copyright: (c) 2015 - 2017 Stichting Koppeltaal
:license: AGPL, see `LICENSE.md` for more details.
"""

import pkg_resources
import zope.interface

NAMESPACE = 'http://ggz.koppeltaal.nl/fhir/Koppeltaal/'

SOFTWARE = 'Koppeltaal python adapter'
VERSION = pkg_resources.get_distribution('koppeltaal').version

OTHER_URL = '/FHIR/Koppeltaal/Other'
ACTIVITY_DEFINITION_URL = '/FHIR/Koppeltaal/Other/_search'
MESSAGE_HEADER_URL = '/FHIR/Koppeltaal/MessageHeader/_search'
METADATA_URL = '/FHIR/Koppeltaal/metadata'
MAILBOX_URL = '/FHIR/Koppeltaal/Mailbox'
OAUTH_AUTHORIZE_URL = 'OAuth2/Koppeltaal/Authorize'
OAUTH_LAUNCH_URL = '/OAuth2/Koppeltaal/Launch'
OAUTH_TOKEN_URL = '/OAuth2/Koppeltaal/Token'

TIMEOUT = 60


class KoppeltaalError(ValueError):
    """Generic koppeltaal error.
    """


class TransportError(KoppeltaalError):
    """There was an error related to the transport.
    """


class ConnectionError(TransportError):
    """There was a connection error from the transport.
    """


class ResponseError(TransportError):
    """The transport got a response but it was an error.
    """

    def __init__(self, response):
        super(ResponseError, self).__init__(response)
        self.response = response

    def __str__(self):
        return "{}: {}.".format(self.__class__.__name__, str(self.response))


class OperationOutcomeError(KoppeltaalError):
    """A non 200 status with an operation outcome was returned by the server.
    """

    def __init__(self, outcome):
        super(OperationOutcomeError, self).__init__(outcome)
        self.outcome = outcome

    def __str__(self):
        issues = []
        for issue in self.outcome.issue:
            issues.append('type: {}, resource: {}, details: {}'.format(
                issue.type, issue.resource.fhir_link, issue.details))
        return "{}: outcome issue(s): {}.".format(
            self.__class__.__name__, issues)


class MessageResponseError(KoppeltaalError):
    """The response in the message header reported an error.
    """

    def __init__(self, message):
        super(MessageResponseError, self).__init__(message)
        self.message = message

    def __str__(self):
        return "{}: {}.".format(self.__class__.__name__, str(self.message))


class DummyError(KoppeltaalError):
    """Error generated when you ask the dummy connector to perform an action.
    """

    def __str__(self):
        return "{}: connector in dummy mode.".format(
            self.__class__.__name__)


class InvalidBundle(KoppeltaalError):
    """A malformed bundle was obtain from the koppeltaal server.
    """


class InvalidValue(KoppeltaalError):
    """A value does not match the definition for its corresponding field
    in the koppeltaal definitions.
    """

    def __init__(self, field, value=None):
        self.field = field
        self.value = value

    def __str__(self):
        python_name = self.field.getName()
        extended = python_name != self.field.name
        if self.field.interface is not None:
            python_name = self.field.interface.getName() + '.' + python_name
        if extended:
            return "{}: invalid value '{}' for '{}' (FHIR name: '{}').".format(
                self.__class__.__name__, self.value,
                python_name, self.field.name)
        return "{}: invalid value '{}' for '{}'.".format(
            self.__class__.__name__, self.value, python_name)


class InvalidReference(InvalidValue):
    """A value that is not a valid reference.
    """

    field = None

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return "invalid reference '{}'.".format(
            self.value.__class__.__name__)


class InvalidCode(InvalidValue):
    """A code appear that is not expected in its definition.
    """

    def __str__(self):
        if self.value:
            return "{}: '{}' not in '{}'.".format(
                self.__class__.__name__,
                self.value,
                self.field.system)
        return "{}: code is missing.".format(self.__class__.__name__)


class InvalidSystem(InvalidCode):
    """The system mentioned by the code is not allowed/supported here.
    """

    def __str__(self):
        if self.value:
            return "{}: system '{}' is not supported.".format(
                self.__class__.__name__,
                self.value)
        return "{}: system is missing.".format(self.__class__.__name__)


class InvalidResource(InvalidValue):
    """A resource type mismatch expected in the koppeltaal definitions.
    """

    def __str__(self):
        if self.field is not None:
            return "{}: expected '{}' resource type.".format(
                self.__class__.__name__,
                self.field.__class__.__name__)
        return "{}: unknown resource type.".format(
            self.__class__.__name__)


class RequiredMissing(InvalidValue):
    """A field is required in the koppeltaal definitions but its value is
    missing.
    """

    def __str__(self):
        return "{}: '{}' required but missing.".format(
            self.__class__.__name__,
            self.field.name)


class IFHIRResource(zope.interface.Interface):
    """A resource that can be sent to the koppeltaal server.
    """

    fhir_link = zope.interface.Attribute(
        'Link to resource containing resource type, id and version')


class IBrokenFHIRResource(IFHIRResource):
    """A resource that was broken.
    """

    error = zope.interface.Attribute(
        'Error')

    payload = zope.interface.Attribute(
        'Resource payload')


class IReferredFHIRResource(IFHIRResource):
    """A resource that is referred but missing from the resource payload
    or bundle.
    """

    display = zope.interface.Attribute(
        'Display value for this resource.')


class IIdentifiedFHIRResource(IFHIRResource):
    """A resource that can be identified with a self link.
    """


class IIntegration(zope.interface.Interface):
    """Personalizing of connector settings and behavior.
    """

    name = zope.interface.Attribute('application name using the connector')

    url = zope.interface.Attribute('fhir base URL for generated resources')

    client_id = zope.interface.Attribute(
        'Application identifier for the domain')

    client_secret = zope.interface.Attribute(
        'OAUTH secret for this application client')

    def transaction_hook(commit_function, message):
        """Optional hook to integrate sending back a message into a
        transaction system.
        """

    def model_id(model):
        """Return a (unique) id for the given `model`. Should stay the same id
        if called again with the same model.
        """

    def link(model, resource_type):
        """Return fhir URL for `model` which is a `resource_type`.
        """


class IUpdate(zope.interface.Interface):
    """Update retrieved from a message via the connector. It can be used a
    context manager to automatically update the processing status of
    the message.
    """
    message = zope.interface.Attribute('message related to the update')

    data = zope.interface.Attribute('focal resource of the message')

    patient = zope.interface.Attribute('patient related to the message')

    def success():
        """Mark the message as succeed.
        """

    def fail(exception="FAILED"):
        """Mark the message as failed, with the given exception message.
        """

    def postpone():
        """Do nothing. The message will stay in claim an eventually revert
        back to new.
        """


class IConnector(zope.interface.Interface):
    """Connector to interact with the koppeltaal server.
    """

    transport = zope.interface.Attribute('transport to access the server')

    domain = zope.interface.Attribute('domain')

    integration = zope.interface.Attribute('integration for this connector')

    def metadata():
        """Return the conformance statement.
        """

    def activities(archived=False):
        """Return a list of activity definitions.
        """

    def activity(identifier, archived=False):
        """Return a specific activity definition identified by `identifier` or
        None.
        """

    def launch(careplan, user=None, activity_identifier=None):
        """Retrieve launch URL for an activity in a careplan for a give user..
        """

    def launch_from_parameters(
            application_id, patient_link, user_link, activity_identifier):
        """Retrieve launch URL out of all required parameters. Those
        parameters can be extracted from a careplan.
        """

    def updates():
        """Iterate over the available new messages in the mailbox for
        processing.
        """

    def search(message_id=None, event=None, status=None, patient=None):
        """Return a list of messages matching the given criteria.
        """

    def send(event, data, patient):
        """Send an update about event with data for patient.
        """

    def close():
        """Close any connection left to the server.
        """
